"""Represents Components and their Parameters."""

from dataclasses import dataclass, field
from http import HTTPStatus
from typing import Any

from froeling import endpoints
from froeling.datamodels.generics import TimeWindowDay
from froeling.exceptions import NetworkError
from froeling.session import Session


class Component:
    """Represents a facility component.

    A `Component` may be partially or fully populated depending on
    how it was fetched.

    - `get_component` methods only set `facility_id` and `component_id`.
    - `Facility.get_components` populates overview data such as
      `display_name`, `display_category`, `standard_name`, `type`,
      and `sub_type`.
    - Call `Component.update()` to load all available values, including
      parameters and other detailed fields.

    Attributes:
        facility_id (int): ID of the facility this component belongs to.
        component_id (str): Unique identifier for the component.
        display_name (str | None): Human-readable name of the component.
        display_category (str | None): High-level category for grouping components.
        standard_name (str | None): Standardized name, if available.
        type (str | None): Component type.
        sub_type (str | None): More specific component subtype.
        time_windows_view (list[TimeWindowDay] | None): Time window data, if fetched.
        picture_url (str | None): URL to a representative image of the component.
        parameters (list[Parameter]): List of associated parameters.
        raw (dict)

    """

    facility_id: int
    component_id: str
    display_name: str | None
    display_category: str | None
    standard_name: str | None
    type: str | None
    sub_type: str | None
    time_windows_view: list[TimeWindowDay] | None
    picture_url: str | None

    parameters: dict[str, 'Parameter']

    raw: dict

    def __init__(self, facility_id: int, component_id: str, session: Session):
        """Initialize a Component with minimal identifying information."""
        self.facility_id = facility_id
        self.component_id = component_id
        self._session = session

        self.time_windows_view = None
        self.picture_url = None
        self.parameters = {}
        self.raw = {}

    @classmethod
    def _from_overview_data(cls, facility_id: int, session: Session, obj: dict) -> 'Component | None':
        """Create a new component and populate it with overview data."""
        component_id = obj.get('componentId')
        if not isinstance(component_id, str):
            return None

        component = cls(facility_id, component_id, session)
        component.display_name = obj.get('displayName')
        component.display_category = obj.get('displayCategory')
        component.standard_name = obj.get('standardName')
        component.type = obj.get('type')
        component.sub_type = obj.get('subType')
        component.raw = obj
        return component

    def __str__(self) -> str:
        """Return a string representation of this component."""
        return f'Component([Facility {self.facility_id}] -> {self.component_id})'

    async def update(self) -> dict[str, 'Parameter']:
        """Update the Parameters of this component."""
        res = await self._session.request(
            'get',
            endpoints.COMPONENT.format(self._session.user_id, self.facility_id, self.component_id),
        )
        self.raw = res
        self.component_id = res.get('componentId')  # This should not be able to change.
        self.display_name = res.get('displayName')
        self.display_category = res.get('displayCategory')
        self.standard_name = res.get('standardName')
        self.type = res.get('type')
        self.sub_type = res.get('subType')
        if res.get('timeWindowsView'):
            self.time_windows_view = TimeWindowDay._from_list(res['timeWindowsView'])  # noqa: SLF001

        #  TODO: Find endpoint that gives all parameters
        topview = res.get('topView')

        parameters: dict[str, dict] = {}
        if topview:
            self.picture_url = topview.get('pictureUrl')
            if 'pictureParams' in topview:
                parameters |= topview.get('pictureParams')
            if 'infoParams' in topview:
                parameters |= topview.get('infoParams')
            if 'configParams' in topview:
                parameters |= topview.get('configParams')
        if 'stateView' in res:
            parameters |= {i['name']: i for i in res.get('stateView')}
        if 'setupView' in res:
            parameters |= {i['name']: i for i in res.get('setupView')}

        self.parameters = Parameter._from_list(list(parameters.values()), self._session, self.facility_id)  # noqa: SLF001
        return self.parameters


@dataclass
class Parameter:
    """Represents a parameter (a value) of a component."""

    session: Session
    facility_id: int

    id: str
    display_name: str | None
    name: str | None
    editable: bool | None
    parameter_type: str | None
    unit: str | None
    value: str | None
    min_val: str | None
    max_val: str | None
    string_list_key_values: dict[str, str] | None

    raw: dict = field(repr=False, default_factory=dict)

    @classmethod
    def _from_dict(cls, obj: dict, session: Session, facility_id: int) -> 'Parameter':
        parameter_id = obj['id']
        display_name = obj.get('displayName')
        name = obj.get('name')
        editable = obj.get('editable')
        parameter_type = obj.get('parameterType')
        unit = obj.get('unit')
        value = obj.get('value')
        min_val = obj.get('minVal')
        max_val = obj.get('maxVal')
        string_list_key_values = obj.get('stringListKeyValues')

        return cls(
            session,
            facility_id,
            parameter_id,
            display_name,
            name,
            editable,
            parameter_type,
            unit,
            value,
            min_val,
            max_val,
            string_list_key_values,
            obj,
        )

    @classmethod
    def _from_list(cls, obj: list[dict], session: Session, facility_id: int) -> dict[str, 'Parameter']:
        """Turn a list of api response dicts into a list of parameter object."""
        return {p.id: p for p in (cls._from_dict(i, session, facility_id) for i in obj)}

    @property
    def display_value(self) -> str:
        """Combine the value with it's unit."""
        if self.string_list_key_values:
            return self.string_list_key_values[str(self.value)]
        if self.unit:
            return f'{self.value} {self.unit}'
        return str(self.value)

    async def set_value(self, value: Any) -> Any | None:
        """Set the value of this parameter.

        Be careful with this, don't change parameters if you don't know what they do.
        You might want to check Parameter.editable together with this.
        Returns None if the value was already the same.
        """
        try:
            return await self.session.request(
                'put',
                endpoints.SET_PARAMETER.format(self.session.user_id, self.facility_id, self.id),
                json={'value': str(value)},
            )
        except NetworkError as e:
            if e.status == HTTPStatus.NOT_MODIFIED:
                return None
            raise



import json
from pathlib import Path

import pytest


@pytest.fixture
def load_json():
    """Load captured JSON responses from tests/responses/."""

    def _loader(name: str):
        path = Path(__file__).parent / 'responses' / name
        with open(path, encoding='utf-8') as f:
            return json.load(f)

    return _loader



"""Datamodels related to the user account."""

from dataclasses import dataclass, field

from froeling.datamodels.generics import Address


@dataclass(frozen=True)
class UserData:
    """Data relating to the user account."""

    email: str | None
    salutation: str | None
    firstname: str | None
    surname: str | None
    address: Address | None
    user_id: int
    lang: str | None
    role: str | None
    active: bool | None
    picture_url: str | None
    facility_count: int | None
    raw: dict = field(repr=False, default_factory=dict)

    @staticmethod
    def _from_dict(obj: dict) -> 'UserData':
        user_data = obj['userData']
        email: str | None = user_data.get('email')
        salutation: str | None = user_data.get('salutation')
        firstname: str | None = user_data.get('firstname')
        surname: str | None = user_data.get('surname')

        address: Address | None = Address._from_dict(user_data['address']) if 'address' in user_data else None  # noqa: SLF001

        user_id: int = user_data.get('userId', -1)
        lang: str | None = obj.get('lang')
        role: str | None = obj.get('role')
        active: bool | None = obj.get('active')
        picture_url: str | None = obj.get('pictureUrl')
        facility_count: int | None = obj.get('facilityCount')
        return UserData(
            email,
            salutation,
            firstname,
            surname,
            address,
            user_id,
            lang,
            role,
            active,
            picture_url,
            facility_count,
            obj,
        )



"""Tests the login process and USER endpoint."""

from unittest.mock import Mock

import pytest
from aioresponses import aioresponses
from froeling import Froeling, endpoints, exceptions


@pytest.mark.asyncio
async def test_login_success(load_json):
    login_data = load_json('login.json')
    token = 'header.eyJ1c2VySWQiOjEyMzR9.signature'

    with aioresponses() as m:
        m.post(
            endpoints.LOGIN,
            status=200,
            payload=login_data,
            headers={'Authorization': token},
        )

        async with Froeling(username='joe', password='pwd') as api:
            userdata = await api.get_userdata()  # should be cached. No new requests
            assert userdata.raw == login_data  # This includes the token.

            assert api.token == token
            assert api.user_id == 1234

            assert userdata.email == 'user@example.com'
            assert userdata.salutation == 'MR'
            assert userdata.firstname == 'James'
            assert userdata.surname == 'Doe'
            assert userdata.lang == 'de'
            assert userdata.role == 'USER'
            assert userdata.active is True
            assert (
                userdata.picture_url
                == 'https://connect-api.froeling.com/aks/connect/v1.0/resources/service/user/12345/picture'
            )
            assert userdata.facility_count is None

            assert userdata.address.street == 'Sesame street'
            assert userdata.address.zip == '12345'
            assert userdata.address.city == 'Somewhere'
            assert userdata.address.country == 'DE'


@pytest.mark.asyncio
async def test_login_failure_raises(load_json):
    with aioresponses() as m:
        m.post(endpoints.LOGIN, status=401, payload=load_json('login_bad_creds.json'))

        with pytest.raises(exceptions.AuthenticationError):
            async with Froeling(username='joe', password='pwd'):
                pass


@pytest.mark.asyncio
async def test_request_auto_reauth(load_json):
    login_data = load_json('login.json')
    user_data = load_json('user.json')
    old_token = 'old.eyJ1c2VySWQiOjEyMzR9.signature'
    new_token = 'new.eyJ1c2VySWQiOjEyMzR9.signature'

    mock_token_callback = Mock()

    with aioresponses() as m:
        m.get(endpoints.USER.format(1234), status=401, body='security check failed')
        m.post(
            endpoints.LOGIN,
            status=200,
            payload=login_data,
            headers={'Authorization': new_token},
        )
        m.get(endpoints.USER.format(1234), status=200, payload=user_data)

        async with Froeling(
            username='joe',
            password='pwd',
            token=old_token,
            auto_reauth=True,
            token_callback=mock_token_callback,
        ) as api:
            d = await api.get_userdata()
            assert d.raw == user_data

        mock_token_callback.assert_called_once_with(new_token)



"""Provides the main API Class."""

import logging
from collections.abc import Callable
from types import TracebackType
from typing import Any

from aiohttp import ClientSession

from froeling import datamodels, endpoints
from froeling.exceptions import FacilityNotFoundError
from froeling.session import Session


class Froeling:
    """The Froeling class provides access to the Fröling API."""

    async def __aenter__(self) -> 'Froeling':
        """Create an API session."""
        try:
            if not self.session.token:
                await self.login()
        except Exception:
            await self.session.close()
            raise
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool | None:
        """End an API session."""
        await self.session.close()
        return None

    def __init__(
        self,
        username: str | None = None,
        password: str | None = None,
        token: str | None = None,
        *,
        auto_reauth: bool = False,
        token_callback: Callable[[str], Any] | None = None,
        language: str = 'en',
        logger: logging.Logger | None = None,
        clientsession: ClientSession | None = None,
    ) -> None:
        """Initialize a Froeling API client instance.

        Either `username` and `password` or a `token` is required.

        Args:
        ----
            username (str | None): Email used to log into your Fröling account.
            password (str | None): Fröling password (not required when using `token`).
            token (str | None): Valid token (not required when using username/password).
            auto_reauth (bool): Automatically fetch a new token if the current one
                expires (requires username and password). Defaults to False.
            token_callback (Callable[[str], Any] | None): Function called when the token
                is renewed (useful for saving the token). Defaults to None.
            language (str): Preferred language for API responses. Defaults to "en".
            logger (logging.Logger | None): Logger for debugging and events.
                Defaults to None.
            clientsession (ClientSession | None): Optional aiohttp session to reuse
                instead of creating a new one. Defaults to None.

        """
        # cached data (does not change often)
        self._userdata: datamodels.UserData | None = None
        self._facilities: dict[int, datamodels.Facility] = {}

        self.session = Session(
            username,
            password,
            token,
            auto_reauth=auto_reauth,
            token_callback=token_callback,
            lang=language,
            logger=logger,
            clientsession=clientsession,
        )
        self._logger = logger or logging.getLogger(__name__)

    async def login(self) -> datamodels.UserData:
        """Log in with the username and password."""
        data = await self.session.login()
        self._userdata = datamodels.UserData._from_dict(data)  # noqa: SLF001
        return self._userdata

    async def close(self) -> None:
        """Close the session."""
        await self.session.close()

    @property
    def user_id(self) -> int | None:
        """The user's id."""
        return self.session.user_id

    @property
    def token(self) -> str | None:
        """The user's token."""
        return self.session.token

    async def _get_userdata(self) -> datamodels.UserData:
        """Fetch userdata (cached)."""
        res = await self.session.request('get', endpoints.USER.format(self.session.user_id))
        return datamodels.UserData._from_dict(res)  # noqa: SLF001

    async def get_userdata(self) -> datamodels.UserData:
        """Get userdata (cached)."""
        if not self._userdata:
            self._userdata = await self._get_userdata()
        return self._userdata

    async def _get_facilities(self) -> list[datamodels.Facility]:
        """Fetch all facilities connected with the account and cache them."""
        res = await self.session.request('get', endpoints.FACILITY.format(self.session.user_id))
        return datamodels.Facility._from_list(res, self.session)  # noqa: SLF001

    async def get_facilities(self) -> list[datamodels.Facility]:
        """Get all cacilities connected with this account (cached)."""
        if not self._facilities:
            facilities = await self._get_facilities()
            self._facilities = {f.facility_id: f for f in facilities}
        return list(self._facilities.values())

    async def get_facility(self, facility_id: int) -> datamodels.Facility:
        """Get a specific facility given it's id (cached)."""
        if facility_id not in self._facilities:
            await self.get_facilities()

        if facility_id not in self._facilities:
            raise FacilityNotFoundError(facility_id)
        return self._facilities[facility_id]

    async def get_notification_count(self) -> int:
        """Fetch the unread notification count."""
        return (await self.session.request('get', endpoints.NOTIFICATION_COUNT.format(self.session.user_id)))[
            'unreadNotifications'
        ]

    async def get_notifications(self) -> list[datamodels.NotificationOverview]:
        """Fetch an overview of all notifications."""
        res = await self.session.request('get', endpoints.NOTIFICATION_LIST.format(self.session.user_id))
        return [datamodels.NotificationOverview(n, self.session) for n in res]

    async def get_notification(self, notification_id: int) -> datamodels.NotificationDetails:
        """Fetch all details for a specific notification."""
        res = await self.session.request('get', endpoints.NOTIFICATION.format(self.session.user_id, notification_id))
        return datamodels.NotificationDetails._from_dict(res)  # noqa: SLF001

    def get_component(self, facility_id: int, component_id: str) -> datamodels.Component:
        """Get a specific component given it's facility_id and component_id.

        Call the update method for this component to populate it's attributes.
        """
        return datamodels.Component(facility_id, component_id, self.session)



"""Exceptions mostly relating to web requests."""

from aiohttp.typedefs import StrOrURL


class AuthenticationError(Exception):
    """Raised for unauthorized requests, failed reauths, or bad credentials."""


class NetworkError(Exception):
    """Raised on unsuccessful HTTP status codes."""

    def __init__(self, msg: str, status: int, url: StrOrURL, res: str) -> None:
        """Initialize a NetworkError.

        Args:
        ----
            msg (str): Short description of the error.
            status (int): HTTP status code returned by the request.
            url (StrOrURL): The requested URL.
            res (str): Raw response body returned by the server.

        """
        super().__init__(f'{msg}: Status: {status}, url: {url}\nResult: {res}')
        self.status = status
        self.url = url


class ParsingError(Exception):
    """Raised when parsing an API response fails.

    Attributes:
        msg (str): Short description of the parsing error.
        doc (str): The raw response text that failed to parse.
        pos (int): The character position in `doc` where the error occurred.
        url (StrOrURL): The URL that was requested.
        lineno (int): Line number of the error position.
        colno (int): Column number of the error position.

    """

    def __init__(self, msg: str, doc: str, pos: int, url: StrOrURL) -> None:
        """Initialize a ParsingError.

        Args:
        ----
            msg (str): Description of the parsing error.
            doc (str): The raw response text.
            pos (int): Character position in `doc` where the error occurred.
            url (StrOrURL): The URL that was requested.

        """
        lineno = doc.count('\n', 0, pos) + 1
        colno = pos - doc.rfind('\n', 0, pos)
        errmsg = (
            f'Error while parsing API response while fetching {url}. {msg}: '
            f'line {lineno} column {colno} (char {pos})\n{doc}'
        )

        super().__init__(errmsg)
        self.msg = msg
        self.doc = doc
        self.pos = pos
        self.url = url
        self.lineno = lineno
        self.colno = colno


class FacilityNotFoundError(Exception):
    """Raised when a requested facility does not exist."""

    def __init__(self, facility_id: int):
        super().__init__(f'Could not find facility with id {facility_id}.')

        self.facility_id = facility_id



"""Datamodels to represent the API objects in python."""

from froeling.datamodels.component import Component, Parameter
from froeling.datamodels.facility import Facility
from froeling.datamodels.notifications import NotificationDetails, NotificationOverview
from froeling.datamodels.userdata import Address, UserData

__all__ = [
    'UserData',
    'Address',
    'NotificationOverview',
    'NotificationDetails',
    'Facility',
    'Component',
    'Parameter',
]



import asyncio
import logging
from dotenv import load_dotenv
import os

from froeling import Froeling

load_dotenv()
username = os.getenv('USERNAME')
password = os.getenv('PASSWORD')
token = os.getenv('TOKEN')

if not (username and password):
    username = input('Username (E-Mail): ')
    password = input('Password         : ')


def print_new_token(token):  # Gets executed when a new token was created (useful for storing the token for next time the program is run)
    print(f'The new token is: {token}')


async def main():
    # You can use an external logger
    logging.basicConfig(level=logging.INFO)

    # When only using token, auto reauth is not available
    async with Froeling(
        username,
        password,
        token=token,
        auto_reauth=True,
        language='en',
        token_callback=print_new_token,
    ) as client:
        for notification in (await client.get_notifications())[:3]:  # Fetch notifications
            await notification.info()  # Load more information about one of the notifications
            print(f'\n[Notification {notification.id}] Subject: {notification.subject}\n{notification.details.body}\n\n')

        facility = (await client.get_facilities())[0]  # Get a list of all facilities
        print(facility)

        # You can see the raw json data corresponding to a datamodel with the .raw attribute.
        print(facility.raw)

        # Get all components of the facility
        example_component = (await facility.get_components())[0]
        print(example_component)

        await example_component.update()  # Get more information about the component. This includes the parameters.
        print(f'{example_component.type} {example_component.sub_type}: {example_component.display_name} \n{"_" * 20}')
        for parameter in (example_component.parameters.values()):  # Loop over all parameters of the component
            print(parameter.display_name, ':', parameter.display_value)

        # You can directly reference a component of a facility by its id
        example_component2 = facility.get_component('1_100')
        await example_component2.update()  # The update method is required to fully populate the component's data.
        print(f'\n\nExample Component: {example_component2.display_name}')

        param = example_component2.parameters.get('7_28')
        if param:
            print(f'Value was {param.display_value}')
            print(f'Setting {param.display_name} to 80')
            # await param.set_value(80) # This changes a live system parameter. Uncomment only if you understand the effect.

        # If you know the facilityId and component_id, you can get the component like this. The data won't be populated.
        client.get_component(facility.facility_id, '1_100')


asyncio.run(main())



"""Datamodels to represent Notifications and related objects."""

import datetime
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from froeling.session import Session

from froeling import endpoints


class NotificationOverview:
    """Stores basic data of a notification."""

    id: int | None
    subject: str | None
    unread: bool | None
    date: datetime.date | None
    error_id: int | None
    type: str | None
    """Known Values: "ERROR", "INFO", "WARNING", "ALARM" """
    facility_id: int | None
    facility_name: str | None
    raw: dict

    details: 'NotificationDetails'

    def __init__(self, data: dict, session: 'Session') -> None:
        """Create a new NotificationOverview."""
        self.session = session
        self._set_data(data)

    def _set_data(self, data: dict) -> None:
        self.raw = data

        self.id = data.get('id')
        self.subject = data.get('subject')
        self.unread = data.get('unread')
        date_str = data.get('notificationDate')
        if isinstance(date_str, str):
            self.date = datetime.datetime.fromisoformat(date_str)
        else:
            self.date = None
        self.error_id = data.get('errorId')
        self.type = data.get('notificationType')
        self.facility_id = data.get('facilityId')
        self.facility_name = data.get('facilityName')

    async def info(self) -> 'NotificationDetails':
        """Get additional information about this notification."""
        res = await self.session.request('get', endpoints.NOTIFICATION.format(self.session.user_id, self.id))
        self.details = NotificationDetails._from_dict(res)  # noqa: SLF001
        return self.details


@dataclass
class NotificationDetails(NotificationOverview):
    """Stores all data related to a notification."""

    body: str | None
    sms: bool | None
    mail: bool | None
    push: bool | None
    notification_submission_state_dto: list['NotificationSubmissionState'] | None
    error_solutions: list['NotificationErrorSolution'] | None

    @classmethod
    def _from_dict(cls, obj: dict) -> 'NotificationDetails':
        body = obj.get('body')
        sms = obj.get('sms')
        mail = obj.get('mail')
        push = obj.get('push')
        submission_state = None
        if 'notificationSubmissionStateDto' in obj:
            submission_state = NotificationSubmissionState._from_list(obj['notificationSubmissionStateDto'])  # noqa: SLF001
        error_solutions = None
        if 'errorSolutions' in obj:
            error_solutions = NotificationErrorSolution._from_list(obj['errorSolutions'])  # noqa: SLF001
        notification_details_object = cls(body, sms, mail, push, submission_state, error_solutions)
        notification_details_object._set_data(obj)  # noqa: SLF001
        return notification_details_object


@dataclass
class NotificationSubmissionState:
    """Submission state of a notification."""

    id: int | None
    recipient: str | None
    type: str | None
    submitted_to: str | None
    submission_result: str | None
    raw: dict = field(repr=False, default_factory=dict)

    @classmethod
    def _from_dict(cls, obj: dict) -> 'NotificationSubmissionState':
        notification_id = obj.get('id')
        recipient = obj.get('recipient')
        notification_type = obj.get('type')
        """Known values: "EMAIL", "TOKEN" """
        submitted_to = obj.get('submittedTo')
        submission_result = obj.get('submissionResult')

        return NotificationSubmissionState(
            notification_id,
            recipient,
            notification_type,
            submitted_to,
            submission_result,
            obj,
        )

    @classmethod
    def _from_list(cls, obj: list[dict]) -> list['NotificationSubmissionState']:
        return [cls._from_dict(i) for i in obj]


@dataclass
class NotificationErrorSolution:
    """Reasons for why the error might occur and steps to take to resolve it."""

    error_reason: str | None
    error_solution: str | None

    @classmethod
    def _from_list(cls, obj: list[dict]) -> list['NotificationErrorSolution']:
        return [cls(i['errorReason'], i['errorSolution']) for i in obj]



"""Test notifications."""

import pytest
from aioresponses import aioresponses
import datetime
from froeling import Froeling, endpoints
from froeling.datamodels.notifications import NotificationSubmissionState


@pytest.mark.asyncio
async def test_get_notification_count(load_json):
    notification_count_data = load_json('notification_count.json')

    token = 'header.eyJ1c2VySWQiOjEyMzR9.signature'

    with aioresponses() as m:
        m.get(
            endpoints.NOTIFICATION_COUNT.format(1234),
            status=200,
            payload=notification_count_data,
        )

        async with Froeling(token=token) as api:
            count = await api.get_notification_count()
            assert count == 123


@pytest.mark.asyncio
async def test_get_notifications(load_json):
    notification_list_data = load_json('notification_list.json')

    token = 'header.eyJ1c2VySWQiOjEyMzR9.signature'

    date = datetime.datetime(
        2025, 1, 2, 12, 34, 56, 789000, tzinfo=datetime.timezone.utc
    )

    with aioresponses() as m:
        m.get(
            endpoints.NOTIFICATION_LIST.format(1234),
            status=200,
            payload=notification_list_data,
        )

        async with Froeling(token=token) as api:
            notifications = await api.get_notifications()
            assert len(notifications) == 3
            for i, n in enumerate(notifications):
                assert n.raw == notification_list_data[i]

                assert n.id == (i + 1) * 10000000 + 123456
                assert n.subject == f'Subject {i + 1}'
                assert n.unread == (False, True, None)[i]
                assert n.date == date
                assert n.error_id == (None, 0, 404)[i]
                assert n.type == ('INFO', 'ALARM', 'ERROR')[i]
                assert n.facility_id == (12345, None, None)[i]
                assert n.facility_name == ('F1', None, None)[i]


@pytest.mark.asyncio
async def test_get_notification_info(load_json):
    notification_list_data = load_json('notification_list.json')
    notification_data = load_json('notification.json')

    token = 'header.eyJ1c2VySWQiOjEyMzR9.signature'

    date = datetime.datetime(
        2025, 1, 2, 12, 34, 56, 789000, tzinfo=datetime.timezone.utc
    )

    with aioresponses() as m:
        m.get(
            endpoints.NOTIFICATION_LIST.format(1234),
            status=200,
            payload=notification_list_data,
        )
        m.get(
            endpoints.NOTIFICATION.format(1234, 10123456),
            status=200,
            payload=notification_data,
        )
        m.get(
            endpoints.NOTIFICATION.format(1234, 10123456),
            status=200,
            payload=notification_data,
        )

        async with Froeling(token=token) as api:
            notification = (await api.get_notifications())[0]
            notification_details = await notification.info()
            assert notification.details == notification_details

            notification_details_2 = await api.get_notification(10123456)
            assert notification_details == notification_details_2

            d = notification_details
            assert d.raw == notification_data
            assert d.id == 10123456
            assert d.subject == 'Subject 1'
            assert d.body == 'Title\r\ntext'
            assert not d.sms
            assert d.mail
            assert d.push
            assert not d.unread
            assert d.date == date
            assert d.type == 'INFO'
            assert d.facility_id == 12345
            assert d.facility_name == 'F1'
            assert len(d.notification_submission_state_dto) == 2
            s1, s2 = d.notification_submission_state_dto

            assert isinstance(s1, NotificationSubmissionState)
            assert s1.raw == notification_data["notificationSubmissionStateDto"][0]
            assert s1.id == 12345678
            assert s1.recipient == 'joe@example.com'
            assert s1.type == 'EMAIL'
            assert s1.submitted_to == 'joe@example.com'
            assert s1.submission_result == 'SUCCESS'

            assert isinstance(s2, NotificationSubmissionState)
            assert s2.raw == notification_data["notificationSubmissionStateDto"][1]
            assert s2.id == 12345679
            assert s2.recipient == 'sometoken'
            assert s2.type == 'TOKEN'
            assert s2.submitted_to == 'joe@example.com'
            assert s2.submission_result == 'SUCCESS'


# TODO: Test NotificationErrorSolution



"""Generic datamodels used in multiple places/endpoints."""

from dataclasses import dataclass, field
from enum import Enum


@dataclass(frozen=True)
class Address:
    """Represents a physical mailing address.

    Attributes:
        street (str): Street name and number.
        zip (int): Postal code.
        city (str): City name.
        country (str): Country name.

    """

    street: str | None
    zip: int | None
    city: str | None
    country: str | None
    raw: dict = field(repr=False, default_factory=dict)

    @staticmethod
    def _from_dict(obj: dict) -> 'Address':
        street = obj.get('street')
        zipcode = obj.get('zip')
        city = obj.get('city')
        country = obj.get('country')
        return Address(street, zipcode, city, country, obj)


class Weekday(Enum):
    """Enumeration of the days of the week."""

    MONDAY = 'MONDAY'
    TUESDAY = 'TUESDAY'
    WEDNESDAY = 'WEDNESDAY'
    THURSDAY = 'THURSDAY'
    FRIDAY = 'FRIDAY'
    SATURDAY = 'SATURDAY'
    SUNDAY = 'SUNDAY'


@dataclass
class TimeWindowDay:
    """Represents the time window schedule for a single day of the week.

    Attributes:
        id (int): Unique identifier for the day entry.
        weekday (Weekday): The day of the week.
        phases (list[TimeWindowPhase]): List of time phases for this day.

    """

    id: int
    weekday: Weekday
    phases: list['TimeWindowPhase']
    raw: dict = field(repr=False, default_factory=dict)

    @classmethod
    def _from_dict(cls, obj: dict) -> 'TimeWindowDay':
        _id = obj['id']
        weekday = Weekday(obj['weekDay'])
        phases = TimeWindowPhase._from_list(obj['phases'])  # noqa: SLF001

        return cls(_id, weekday, phases, obj)

    @classmethod
    def _from_list(cls, obj: list) -> list['TimeWindowDay']:
        return [cls._from_dict(i) for i in obj]


@dataclass
class TimeWindowPhase:
    """Represents a time phase within a single day.

    Attributes:
        start_hour (int): Hour when the phase starts (0-23).
        start_minute (int): Minute when the phase starts (0-59).
        end_hour (int): Hour when the phase ends (0-23).
        end_minute (int): Minute when the phase ends (0-59).

    """

    start_hour: int
    start_minute: int
    end_hour: int
    end_minute: int
    raw: dict = field(repr=False, default_factory=dict)

    @classmethod
    def _from_dict(cls, obj: dict) -> 'TimeWindowPhase':
        sh = obj['startHour']
        sm = obj['startMinute']
        eh = obj['endHour']
        em = obj['endMinute']

        return cls(sh, sm, eh, em, obj)

    @classmethod
    def _from_list(cls, obj: list) -> list['TimeWindowPhase']:
        return [cls._from_dict(i) for i in obj]
